hooksecurefunc("ObjectiveTracker_Initialize", function()
	ObjectiveTrackerFrame.MODULES = {	SCENARIO_CONTENT_TRACKER_MODULE,
						AUTO_QUEST_POPUP_TRACKER_MODULE,
						QUEST_TRACKER_MODULE,
						BONUS_OBJECTIVE_TRACKER_MODULE,
						WORLD_QUEST_TRACKER_MODULE,
						ACHIEVEMENT_TRACKER_MODULE,
	};
	ObjectiveTrackerFrame.MODULES_UI_ORDER = {	SCENARIO_CONTENT_TRACKER_MODULE,
								AUTO_QUEST_POPUP_TRACKER_MODULE,
								BONUS_OBJECTIVE_TRACKER_MODULE,
								WORLD_QUEST_TRACKER_MODULE,
								QUEST_TRACKER_MODULE,
								ACHIEVEMENT_TRACKER_MODULE,
	};
end)

local function Merge(source, destination, looped, blackList)
	destination = destination or {}
	if not source then
		return destination
	end

	for key, value in pairs(CopyTable(source)) do
		if (type(value) == 'table') then
			if type(key) == "number" then
				local index = #destination + 1
				destination[index] = Merge(value, destination[index], true, blackList)
			else
				destination[key] = Merge(value, destination[key], true, blackList)
			end
		else
			if (not looped) and (blackList and blackList[key]) then

			else
				destination[key] = value
			end
		end
	end
	return destination
end

local handler = {}
local it

handler.defaults = {
			height = 420,
			enableScroll = true,
			achieveheight = 160,
			proximitySort = true,
			alwaysShowAchieves = true,
			point = "Right",
			x = -52,
			y = 30,
			pad = 0,
			scale = 1,
		}

--frame creation
do
	function handler:Create()
		local panel = handler.self or CreateFrame("Frame", "ObjectiveContainer", UIParent)
		handler.self = panel
		panel:SetPoint("TopLeft", MinimapCluster, "BottomLeft", -120, 0)
		panel:SetSize(273, 420)
		panel.QuestList = {}
		Merge(handler, panel)
		return panel
	end

	function handler:InitializeController(parent, conf)
		local panel = handler:Create()
		panel:SetParent(parent)
		panel:ClearAllPoints()
		panel:SetPoint('TopLeft', 0, 0)
		panel:SetPoint('BottomRight', 0, 0)
		panel.parent = parent
		if Dom_Quest and Dom_Quest.noDominos then
			parent.sets = Dom_Quest
			conf = Dom_Quest
			Dom_Quest.noDominos = nil
		end
		panel.sets = conf
		Dom_Quest = conf
		panel:Initialize()
		
		function panel:ParentLayout()
			parent:Layout()
		end
		return panel
	end

	function handler:Initialize()
		local panel = handler:Create()
		if not Dom_Quest then
			Dom_Quest = Dom_Quest or handler.defaults
			Dom_Quest.noDominos = true
		end
		panel.sets = panel.sets or Dom_Quest
		panel:EnableStoredFunctions()
		--corrects bug where all quests display an extra empty line of text.
		for i, mod in pairs({QUEST_TRACKER_MODULE, ACHIEVEMENT_TRACKER_MODULE}) do
			hooksecurefunc(mod, "SetBlockHeader", function()
				panel:RepositionTracker()
			 end)
		end
		panel:HookMinimizeButton()
		panel.scrollValue = panel.scrollValue or 0
		panel:GetQuests():HookScript("OnEvent", function()
			panel:ScrollQuests()
			panel:ScrollAchieves()
		end)
		panel:ScrollQuests()
		panel:ScrollAchieves()
		panel:RegisterEvent("ZONE_CHANGED")
		panel:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		panel:RegisterEvent("ZONE_CHANGED_INDOORS")
		panel:RegisterEvent("ADVENTURE_MAP_QUEST_UPDATE")
		panel:SetScript("OnEvent", panel.UpdateList)
		panel:Layout()
		panel:RepositionCollapse()
		panel:UpdateAlwaysShowAchievements()
		hooksecurefunc("QuestObjectiveSetupBlockButton_AddRightButton", function(block, button, initialAnchorOffsets)
			self:AdjustBlock(block, button)
		end)
		panel:UpdateAchievementHeight()
		return panel
	end
end

--objects
do
	function handler:Get()
		return handler.self
	end

	function handler:GetQuests()
		return ObjectiveTrackerFrame
	end

	function handler:GetQuests_Mini()
		return ObjectiveTrackerFrame.HeaderMenu.MinimizeButton
	end

	function handler:GetAchievements()
		return AchievementTrackerFrame
	end

	function handler:GetAchieves_Mini()
		return AchievementTrackerFrame.HeaderMenu.MinimizeButton
	end
end

--layout objects
do
	function handler:Layout()
		self:RepositionTracker()
		self:UpdateAlwaysShowAchievements()
		self:RepositionCollapse()
		self:AdjustAppearance()
		self:ScrollQuests()
		self:ScrollAchieves()
		self:Reposition()
		self:UpdateList()
		
		if self.ParentLayout then
			self:ParentLayout()
		end

	end

	function handler:RepositionTracker()
		local quests = self:GetQuests()
		local AT = self:GetAchievements()


		if not quests.set then
			quests.set = true
			quests:ClearAllPoints()
			AT:SetParent(self)
			AT:ClearAllPoints()
		end
		quests:SetPoint('TopLeft', self, 30, 0)
		quests:SetPoint('TopRight', self, 0, 0)

		local achieveheight = self:GetSetting("achieveheight")
		local height = self:GetSetting("height")

		AT:SetPoint("BottomLeft", self, "BottomLeft", 30, 0)
		AT:SetPoint("BottomRight", self, "BottomRight", 0, 0)
		AT:SetHeight(achieveheight)

		
		if self:GetSetting("alwaysShowAchieves") == true then
			AT:Show()
			ObjectiveTrackerBlocksFrame.AchievementHeader:Hide()
			
			quests:SetHeight(height)

		else
			AT:Hide()
			ObjectiveTrackerBlocksFrame.AchievementHeader:Show()
			quests:SetHeight(height + achieveheight)

		end
	
		self:SetHeight(height + achieveheight)
	end

	function handler:UpdateAlwaysShowAchievements()
		if self:GetSetting("alwaysShowAchieves") == true then
			local displayOnlyArena = ArenaEnemyFrames and ArenaEnemyFrames:IsShown() and (instanceType == "arena");
			local trackedAchievements = { self:GetFunction("GetTrackedAchievements", true)() };
			for i = 1, #trackedAchievements do
				local achievementID = trackedAchievements[i];
				local block = ACHIEVEMENT_TRACKER_MODULE:GetBlock(achievementID)
				if block then
					block:Hide()

					ACHIEVEMENT_TRACKER_MODULE:SetBlockHeader(block, "")
					ACHIEVEMENT_TRACKER_MODULE:FreeUnusedLines(block)
				end
			end
			if ObjectiveTrackerFrame.MODULES and #ObjectiveTrackerFrame.MODULES == 6 then
				tremove(ObjectiveTrackerFrame.MODULES, 6)
				tremove(ObjectiveTrackerFrame.MODULES_UI_ORDER, 6)
			end
		else
			if ObjectiveTrackerFrame.MODULES and #ObjectiveTrackerFrame.MODULES < 6 then
				tinsert(ObjectiveTrackerFrame.MODULES, 6, ACHIEVEMENT_TRACKER_MODULE)
				tinsert(ObjectiveTrackerFrame.MODULES_UI_ORDER, 6, ACHIEVEMENT_TRACKER_MODULE)
			end
		end
	end

	function handler:RepositionCollapse()
		local quest = self:GetQuests_Mini()
		quest:ClearAllPoints()

		local achieve = self:GetAchieves_Mini()
		achieve:ClearAllPoints()

		local AT = self:GetAchievements()

		local alwaysShow = self:GetSetting("alwaysShowAchieves")

		local anchor, point = achieve, "TopRight"

		if alwaysShow ~= true then
			achieve:Hide()
			if self.mover and self.mover.green then
				self.mover.green:Hide()

			end
			anchor = self
			point = "BottomRight"
		else
			achieve:Show()
			if self.mover and self.mover.green then
				self.mover.green:Show()
			end
		end


		if self:GetSetting("revCollapse") then
			if self.miniClicked2 then
				achieve:SetPoint("BottomRight", AT, 0, 0)
			else
				achieve:SetPoint("TopRight", AT, 0, 0)
			end

			if self.miniClicked1 then
				quest:SetPoint("BottomRight", anchor, point, 0, 0)
			else
				quest:SetPoint("TopRight", self, 0, 0)
			end

		else
			achieve:SetPoint("TopRight", AT, 0, 0)
			quest:SetPoint("TopRight", self, 0, 0)
		end


	end

	function handler:AdjustAppearance()
		local watches = {
			QUEST_TRACKER_MODULE = {num = GetNumQuestWatches, info = self:GetFunction("GetQuestWatchInfo", true)},
			WORLD_QUEST_TRACKER_MODULE = {num = GetNumWorldQuestWatches, info = GetWorldQuestWatchInfo},
			ACHIEVEMENT_TRACKER_MODULE = {num = function() local list = {self:GetFunction("GetTrackedAchievements", true)()} return #list end, info = function(i)
				local list = {self:GetFunction("GetTrackedAchievements")()}
					local achievementID = list[i]
				return GetAchievementInfo(achievementID)
			end},
		}

		for name, functions in pairs(watches) do
			for i = 1, functions.num() do
				local id = functions.info(i)
				if id then
					local block = _G[name]:GetExistingBlock(id)
					if block then
						self:AdjustBlock(block,  name, id)
					end
				end
			end
		end
	end

	function handler:ToggleCollapse(buttonIndex)
		self["miniClicked"..buttonIndex] = not self["miniClicked"..buttonIndex]
		self:RepositionCollapse()
	end
end

--utility functions
do
	function handler:GetSetting(setting)
		if not self.sets then
			return self.defaults[setting]
		else
			return self.sets[setting]
		end
	end

	function handler:ToggleSetting(setting)
		self.sets[setting] = not self.sets[setting]
		if self.sets[setting] == false then
			self.sets[setting] = nil
		end
		self:Layout()
	end

	function handler:UpdateSetting(setting, value)
		if self.sets[setting] then
			self.sets[setting] = value
		end
		self:Layout()
	end

	function handler:AdjustSetting(setting, value)
		if self.sets[setting] then
			self.sets[setting] = self.sets[setting] + value
		end
		self:Layout()
	end

	function handler:UpdateList()
		ObjectiveTracker_Update()
		AchievementTracker_Update()
	end

	function handler:ScrollQuests(delta)
		if (not self:GetSetting("enableScroll")) or (not delta) then
			self.scrollValue = 0
		else
			self.scrollValue = self.scrollValue + delta
		end
		local _min, _max = 0, GetNumQuestWatches() - 1
		if self.scrollValue > _max then
			self.scrollValue = _min
		end
		if self.scrollValue < _min then
			self.scrollValue = _max
		end
		self:UpdateList()
	end

	function handler:ScrollAchieves(delta)
		if (not self:GetSetting("enableScroll")) or (not delta) then
			self.scrollAchieveValue = 0
		else
			self.scrollAchieveValue = self.scrollAchieveValue + delta
		end

		local list = {self:Get():GetFunction("GetTrackedAchievements", true)()}

		local _min, _max = 0, #list - 1
		if self.scrollAchieveValue > _max then
			self.scrollAchieveValue = _min
		end
		if self.scrollAchieveValue < _min then
			self.scrollAchieveValue = _max
		end
		self:UpdateList()
	end

	function handler:UpdateAchievementHeight()
		self:RepositionTracker()
		self:UpdateList()
	end

	local buttons = {
		"itemButton",
		"groupFinderButton",
		"rightButton",
	}

	function handler:AdjustBlock(block)
		if ( block ) then
			for i, line in pairs(block.lines) do
				line.Text:SetHeight(28)
			end
			for i, itemButton in pairs(buttons)do
				itemButton = block[itemButton]
				if ( itemButton ) then
					itemButton:SetMovable(true)
					itemButton:SetUserPlaced(true)
					itemButton:ClearAllPoints()
					if self:GetSetting("onLeft") then
						itemButton:SetPoint("TopLeft", block, -50, 0)
					else
						itemButton:SetPoint("TopRight", block, -2, 0)
					end
				end
			end
		end
	end

	function handler:SetMoving(move)
		self.moving = not self.moving
		self:SetMovable(true)
		self:SetClampedToScreen(true)
		if self.moving then
			if not self.mover then
				self.mover = CreateFrame("Frame", nil, self)
				self.mover:SetAllPoints(self)
				local blue = self.mover:CreateTexture(nil, 'TOOLTIP')
				blue:SetPoint("TopLeft", self:GetQuests(), -30,0)
				blue:SetPoint("BottomRight", self:GetAchievements(), "TopRight", 0,0)
				blue:SetColorTexture(0,0,1, .25)
				local green = self.mover:CreateTexture(nil, 'TOOLTIP')
				green:SetPoint("TopLeft", self:GetAchievements(), -30,0)
				green:SetPoint("BottomRight", self:GetAchievements(), 0,0)
				green:SetColorTexture(0,1,0, .25)
				self.mover.green = green
				self.mover:SetScript("OnMouseDown", function(_, button)
					if button == "LeftButton" then
						self:StartMoving()
					else
						self:SetMoving(nil)
					end
				end)
				self.mover:SetScript("OnMouseUp", function()
					self:StopMovingOrSizing()
					self:SaveRelativeFramePosition()
				end)
			end


			self.mover:Show()

		elseif self.mover then
			self.mover:Hide()
		end

		if self:GetSetting("alwaysShowAchieves") ~= true then
			if self.mover and self.mover.green then
				self.mover.green:Hide()
			end
		else
			if self.mover and self.mover.green then
				self.mover.green:Show()
			end
		end

	end

	function handler:SaveRelativeFramePosition()
		self:SaveFramePosition(self:GetRelativeFramePosition())
	end

	function handler:GetRelativeFramePosition()
		local scale = self:GetScale() or 1
		local left = self:GetLeft() or 0
		local top = self:GetTop() or 0
		local right = self:GetRight() or 0
		local bottom = self:GetBottom() or 0

		local parent = self:GetParent() or _G['UIParent']
		local pwidth = parent:GetWidth() / scale
		local pheight = parent:GetHeight() / scale

		local x , y, point
		if left < (pwidth - right) and left < abs((left+right)/2 - pwidth/2) then
			x = left
			point = 'LEFT'
		elseif (pwidth - right) < abs((left + right)/2 - pwidth/2) then
			x = right - pwidth
			point = 'RIGHT'
		else
			x = (left+right)/2 - pwidth/2
			point = '';
		end

		if bottom < (pheight - top) and bottom < abs((bottom + top)/2 - pheight/2) then
			y = bottom
			point = 'BOTTOM' .. point
		elseif (pheight - top) < abs((bottom + top)/2 - pheight/2) then
			y = top - pheight
			point = 'TOP' .. point
		else
			y = (bottom + top)/2 - pheight/2
		end

		if point == '' then
			point = 'CENTER'
		end

		return point, x, y
	end


	--[[ Position Saving ]]--

	local roundPoint = function(point)
		point = point or 0

		if point > 0 then
			point = floor(point + 0.5)
		else
			point = ceil(point - 0.5)
		end

		return point
	end

	function handler:SetFramePosition(...)
		self:ClearAllPoints()
		self:SetPoint(...)
	end

	function handler:Reposition()
		--self:Rescale()
		self:SetFramePosition(self:GetSavedFramePosition())
	end

	function handler:SaveFramePosition(point, x, y)
		point = point or 'CENTER'
		x = roundPoint(x)
		y = roundPoint(y)

		local sets = self.sets
		sets.point = point ~= 'CENTER' and point or nil
		sets.x = x ~= 0 and x or nil
		sets.y = y ~= 0 and y or nil
		self:SetUserPlaced(true)


		--print(x,y,point)
	end

	function handler:GetSavedFramePosition()
		local sets = self.sets
		local point = sets.point or 'CENTER'
		local x = sets.x or 0
		local y = sets.y or 0

		return point, x, y
	end
end

--menu options
do
	function handler:GetMenuOptions(level, menuValue)
		if level == 1 then
			return {
					{
						text = "Filter By Zone",
						func = function() return self:ToggleSetting("filterByZone") end,
						checked = function() return self:GetSetting("filterByZone") end,
						keepShownOnClick = true,
						dispCondition = function() return true end,
					},
					{
						text = "Hide Group Quests",
						func = function() return self:ToggleSetting("filterByGroup") end,
						checked = function() return self:GetSetting("filterByGroup") end,
						keepShownOnClick = true,
						dispCondition = function() return true end,
					},
					{
						text = "Hide Campaign",
						func = function() return self:ToggleSetting("filterByCampaign") end,
						checked = function() return self:GetSetting("filterByCampaign") end,
						keepShownOnClick = true,
						dispCondition = function() return true end,
					},

					{
						text = "Sort By Distance",
						func = function() return self:ToggleSetting("proximitySort") end,
						checked = function() return self:GetSetting("proximitySort") end,
						keepShownOnClick = true,
						dispCondition = function() return true end,
					},
					{
						text = "Enable Scrolling",
						func = function() return self:ToggleSetting("enableScroll") end,
						checked = function() return self:GetSetting("enableScroll") end,
						keepShownOnClick = true,
						dispCondition = function() return true end,
					},
					{
						text = "Always Show Achievements",
						func = function() return self:ToggleSetting("alwaysShowAchieves") end,
						checked = function() return self:GetSetting("alwaysShowAchieves") end,
						keepShownOnClick = true,
						dispCondition = function() return true end,
					},
					{
						text = "Buttons on Left",
						func = function() return self:ToggleSetting("onLeft") end,
						checked = function() return self:GetSetting("onLeft") end,
						keepShownOnClick = true,
						dispCondition = function() return true end,
					},
					{
						text = "Collapse Down",
						func = function() return self:ToggleSetting("revCollapse") end,
						checked = function() return self:GetSetting("revCollapse") end,
						keepShownOnClick = true,
						dispCondition = function() return true end,
					},
					{
						text = "Set Movable",
						func = function() return self:SetMoving(true) end,
						checked = function() return self.moving end,
						keepShownOnClick = true,
						dispCondition = function() if not Dominos then return true end end,
					},

					{
						text = "Quest Height",
						value = "height",
						keepShownOnClick = true,
						displayMode = "MENU",
						hasArrow = 1,
						--notClickable = true,
						dispCondition = function() if not Dominos then return true end end,
					},
					{
						text = "Achievement Height",
						value = "achieveheight",
						keepShownOnClick = true,
						displayMode = "MENU",
						hasArrow = 1,
						--notClickable = true,
						dispCondition = function() if not Dominos then return true end end,
					},
			}
		end
		if (level == 2) and not Dominos then
			local menu = {}
			for i, number in pairs({-25,-10,-5, -1, "x", 1,5,10,25}) do
				local inf = {}
				if number == "x" then
					inf.text = self:GetSetting(menuValue)
					inf.notClickable = true
				else
					if number > 0 then
						inf.text = "+"..number;
					else
						inf.text = number;
					end
					inf.func = function()
						self:AdjustSetting(menuValue, number)
					end
				end

				inf.notCheckable = true
				inf.keepShownOnClick  = true
				tinsert(menu, inf)
			end
			return menu
		end
	end

	function handler:InitializeMenu(level, ...)
		if level == 1 then
			for i, inf in pairs(self:GetMenuOptions(level)) do
				if inf.dispCondition() then
					local info = UIDropDownMenu_CreateInfo()
					for j, b in pairs(inf) do
						info[j] = b
					end
					UIDropDownMenu_AddButton(info)
				end
			end
		elseif not dominos then
			local inf = self:GetMenuOptions(level, UIDROPDOWNMENU_MENU_VALUE)
			for i, inf in pairs(inf or {}) do
				local info = UIDropDownMenu_CreateInfo()
				for j, b in pairs(inf) do
					info[j] = b
				end
				UIDropDownMenu_AddButton(info,  2)
			end
		end
	end
end

--modify existing functions or frames
do
	local SEAL_QUESTS = {
		[40519] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_VARIAN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal"},
		[43926] = { bgAtlas = "QuestBG-Horde", text = "|cff480404"..QUEST_WARCHIEF_VOLJIN.."|r", sealAtlas = "Quest-Horde-WaxSeal"},
		[47221] = { bgAtlas = "QuestBG-TheHandofFate", },
		[47835] = { bgAtlas = "QuestBG-TheHandofFate", },
		[49929] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_ANDUIN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal" },
		[49930] = { bgAtlas = "QuestBG-Horde", text = "|cff480404"..QUEST_WARCHIEF_SYLVANAS_WINDRUNNER.."|r", sealAtlas = "Quest-Horde-WaxSeal" },
		[50476] = { bgAtlas = "QuestBG-Horde", sealAtlas = "Quest-Horde-WaxSeal" },
		-- BfA start quests
		[46727] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_ANDUIN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal" },
		[50668] = { bgAtlas = "QuestBG-Horde", text = "|cff480404"..QUEST_WARCHIEF_SYLVANAS_WINDRUNNER.."|r", sealAtlas = "Quest-Horde-WaxSeal"},

		[51795] = { bgAtlas = "QuestBG-Alliance" },
		[52058] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_ANDUIN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal"},

		[51796] = { bgAtlas = "QuestBG-Horde" },

		[53372] = { bgAtlas = "QuestBG-Horde", text = "|cff480404"..QUEST_WARCHIEF_SYLVANAS_WINDRUNNER.."|r", sealAtlas = "Quest-Horde-WaxSeal"},
		[53370] = { bgAtlas = "QuestBG-Alliance", text = "|cff042c54"..QUEST_KING_ANDUIN_WRYNN.."|r", sealAtlas = "Quest-Alliance-WaxSeal"},
	}

	local EXCEPTION_QUESTS = {
		[53029] = true,
		[53026] = true,
		[51211] = true,
		[52428] = true,
	}

	function handler:EnableStoredFunctions()
		local funcs
		funcs = {
			-- [[
			ACHIEVEMENT_TRACKER_MODULE ={

												Update =	{
																newFunction = function(...)
																	if self:GetSetting("alwaysShowAchieves") ~= true then
																		self:GetFunction("ACHIEVEMENT_TRACKER_MODULE", "Update", true)(...)
																	end
																end,
															},
												GetBlock =	{
																newFunction = function(...)
																	local block = self:GetFunction("ACHIEVEMENT_TRACKER_MODULE", "GetBlock", true)(...)

																	block.module.SCROLL = block.module.SCROLL or function(step) self:ScrollAchieves(step)  end

																	if block and not block.scrollEnabled then
																		block.scrollEnabled = true
																		block:SetScript("OnMouseWheel", function(_, step)
																			if block.module.SCROLL then
																				block.module.SCROLL(step)
																			end
																		end)
																	end
																	return block
																end,
															},

										},

			DEFAULT_OBJECTIVE_TRACKER_MODULE =	{
												GetBlock = {
													newFunction = function(...)
														local block = self:GetFunction("DEFAULT_OBJECTIVE_TRACKER_MODULE", "GetBlock", true)(...)

														block.module.SCROLL = block.module.SCROLL or function(step) self:ScrollQuests(step)  end
														if block and not block.scrollEnabled then
															block.scrollEnabled = true
															block:SetScript("OnMouseWheel", function(_, step)
																block.module.SCROLL(step)
															end)
														end
														return block
													end,
												},

										},

			_ACHIEVEMENT_TRACKER_MODULE =	{
												GetBlock = {
													newFunction = function(...)
														local block = self:GetFunction("_ACHIEVEMENT_TRACKER_MODULE", "GetBlock", true)(...)

														block.module.SCROLL = block.module.SCROLL or function(step) self:ScrollAchieves(step)  end
														if block and not block.scrollEnabled then
															block.scrollEnabled = true
															block:SetScript("OnMouseWheel", function(_, step)
																block.module.SCROLL(step)
															end)
														end
														return block
													end,
												},

											},

			GetQuestWatchInfo =	{
									newFunction = function(i)
										if (self:GetSetting("hideQuests") == true) then
											return
										end
										self.scrollValue = self.scrollValue or 0
										local scroll = self.scrollValue
										local _min, _max = 1, self:GetFunction("GetNumQuestWatches")()
										local _i = i - scroll
										if _i > _max then
											_i = _min + (_i - _max)
										end
										if _i < _min then
											_i = _max + (_i)
										end
										local info = self.QuestList[_i]
										if info then
											return unpack(info)
										end
									end,
								},
			GetTrackedAchievements =	{
											newFunction = function(...)
												local list = {self:GetFunction("GetTrackedAchievements", true)(...)}
												self.scrollAchieveValue = self.scrollAchieveValue or 0
												local scroll = self.scrollAchieveValue
												local _min, _max = 1, #list
												local rip = {}
												for item, info in pairs(list) do
													local _i = item - scroll
													if _i > _max then
														_i = _min + (_i - _max)
													end
													if _i < _min then
														_i = _max + (_i)
													end
													local f = {id = info, INDEX = _i}
													tinsert(rip, f)
												end
												table.sort(rip, function(a,b) if a.INDEX and b.INDEX then return a.INDEX < b.INDEX end end);
												local r = {}
												for i, b in pairs(rip) do
													r[i] = b.id
												end
												return unpack(r)
											end,
										},

			GetNumQuestWatches =	{
										newFunction = function(...)
											local hasAnyPoi
											local list = self.QuestList
											wipe(list)
											local proximitySort = self:GetSetting("proximitySort")
											for item = 1, self:GetFunction("GetNumQuestWatches", true)() do
												local questInfo = {self:GetFunction("GetQuestWatchInfo", true)(item)};
												if proximitySort and questInfo[3] then
													questInfo.distance = GetDistanceSqToQuest(questInfo[3]) or math.huge
												end
												local shouldInsert = true
												if not questInfo then
													break
												end
												local one, two, three = questInfo[1], questInfo[15], questInfo[16]
												local hasPoi
												if (two == true) or ((three == true)) or ( (C_CampaignInfo.IsCampaignQuest(one) or SEAL_QUESTS[one]))then
													hasPoi = true
													hasAnyPoi = true
												end
												questInfo.hasPoi = hasPoi
												if one then
													if (C_LFGList.CanCreateQuestGroup(one)) and (self:GetSetting("filterByGroup") == true) then
														shouldInsert = nil
													end
													if (C_CampaignInfo.IsCampaignQuest(one) or SEAL_QUESTS[one]) and (self:GetSetting("filterByCampaign") == true) then
														shouldInsert = nil
													end
												end
												if shouldInsert then
													tinsert(list, questInfo)
												end
											end
											local l = {}
											if hasAnyPoi and (self:GetSetting("filterByZone") == true) then
												for i, b in pairs(list) do
													if b.hasPoi == true then
														tinsert(l, b)
													end
												end
												list = l
											end
											if proximitySort then
												table.sort(list, function(a,b) if a.distance and b.distance then return a.distance < b.distance end end);
											end
											return #list or self:GetFunction("GetNumQuestWatches", true)()
										end,
									},
			--]]
		}

		for name, info in pairs(funcs) do
			if info.newFunction then
				info.oldFunction = _G[name]
				_G[name] = info.newFunction
			else
				for functionName, info in pairs(info) do
					info.oldFunction = _G[name][functionName]
					_G[name][functionName] = info.newFunction
				end
			end
		end
		self.StoredFunctions = funcs
	end

	function handler:GetStoredFunction()
		return self.StoredFunctions
	end

	function handler:GetFunction(frameName, functionName, getOriginal)
		local funcs = self:GetStoredFunction()
		local funcInfo = funcs[frameName]

		if type(functionName) ~= "string" then
			getOriginal = functionName
		else
			funcInfo = funcInfo[functionName]
		end

		if getOriginal then
			return funcInfo.oldFunction
		else
			return funcInfo.oldFunction
		end
	end

	function handler:HookMinimizeButton()
		--Toggle display of quest. Leaves tracked Achievements(etc.) visible.
		local dropDown = CreateFrame("Frame", "QuestFilterMenu", self, "UIDropDownMenuTemplate")
		self.dropDown = dropDown

		UIDropDownMenu_Initialize( dropDown, function(_, level, ...) return self:InitializeMenu(level, ...) end, "MENU")
		for i, mini in pairs({self:GetQuests_Mini(), self:GetAchieves_Mini()}) do
			mini:ClearAllPoints()
			mini:SetPoint("TopRight", self, 0, 0)
			mini:SetParent(self)
			local minimizeHandler = CreateFrame("Frame", "DominosQuestMinimizeButtonOverlay"..i, mini)
			minimizeHandler:SetParent(mini)
			minimizeHandler:SetAllPoints(mini)
			minimizeHandler:EnableMouse("AnyUp")
			mini:SetScript("OnHide", function() end)
			minimizeHandler:SetScript("OnMouseDown", function(_, button)
				if button == "RightButton" then
					ToggleDropDownMenu(1, 1, dropDown, minimizeHandler:GetName(), 0, -5)
				else
					mini:Click()
					self:ToggleCollapse(i)
				end
			end)
		end
	end
end

if not Dominos then
	local loader = CreateFrame("Frame")
	loader:RegisterEvent("PLAYER_ENTERING_WORLD")
	loader:SetScript("OnEvent", function()
		handler:Initialize()
		loader:SetScript("OnEvent", nil)
	end)
	return
end

	--[[ Globals ]]--
local AddonName = ...
local Addon = LibStub('AceAddon-3.0'):GetAddon("Dominos")
local KeyBound = LibStub('LibKeyBound-1.0')

--[[ Bar ]]--
local QuestBar = Addon:CreateClass('Frame', Addon.Frame)
QuestBar.master = Addon
_G[...] = QuestBar

do
	local QuestLogController = QuestBar.master:NewModule('QuestLogBar', 'AceEvent-3.0')

	function QuestLogController:Load()
		self.bar = QuestBar:New()
		
		self.bar.panel = handler:InitializeController(self.bar, self.bar.sets)
		self.bar.panel.parent = self.bar
		
		self.bar:Layout()
	end

	function QuestLogController:Unload()
		self:UnregisterAllEvents()

		if self.bar then
			self.bar:Free()
		end
	end

	function QuestLogController:UpdateButtons()
		if InCombatLockdown() then return end
		self.bar:UpdateButtons()
	end

	function QuestBar:New()
		local frame = QuestBar.proto.New(self, 'quest')


		return frame
	end

	function QuestBar:Layout()
		local w = self.panel:GetWidth()
		h = self.sets.height


--		self.panel:Layout()

		local width, height = self.panel:GetSize()

		local pad = self:GetPadding()




		self:SetSize(width +(pad *1.5), height+(pad*1.5))

		self.panel:ClearAllPoints()
		self.panel:SetPoint("Center")
	end

	function QuestBar:GetDefaults()
		return {
			spacing = 0,
			point = "Right",
			x = -52,
			y = 30,
			height = 420,
			scrollStep = 50,
			enableScroll = true,
			pad = 0,
			achieveheight = 160,
			proximitySort = true,
			alwaysShowAchieves = true,
		}
	end

	function QuestBar:ToggleSetting(setting)
		self.sets[setting] = not self.sets[setting]
		if self.sets[setting] == false then
			self.sets[setting] = nil
		end
		self.panel:Layout()
	--	self:Layout()
	end

	function QuestBar:UpdateSetting(setting, value)
		if self.sets[setting] then
			self.sets[setting] = value
		end
		self.panel:Layout()
--		self:Layout()
	end

	
	local L

	local function AddQuestPanel(menu)
		local panel = menu:NewPanel("Quests")
		
		local c = panel:NewCheckButton{
			name = "Zone Filter",
			get = function() return panel.owner.sets.filterByZone end,
			set = function(self, enable)
				panel.owner:ToggleSetting("filterByZone")
			end
		}

		local c = panel:NewCheckButton{
			name = "Hide Group",
			get = function() return panel.owner.sets.filterByGroup end,
			set = function(self, enable)
				panel.owner:ToggleSetting("filterByGroup")
			end
		}

		local c = panel:NewCheckButton{
			name = "Hide Campaign",
			get = function() return panel.owner.sets.filterByCampaign end,
			set = function(self, enable)
				panel.owner:ToggleSetting("filterByCampaign")
			end
		}

		local c = panel:NewCheckButton{
			name = "Sort By Distance",
			get = function() return panel.owner.sets.proximitySort end,
			set = function(self, enable)
				panel.owner:ToggleSetting("proximitySort")
			end
		}

		local c = panel:NewCheckButton{
			name = "Buttons on Left",
			get = function() return panel.owner.sets.onLeft end,
			set = function(self, enable)
				panel.owner:ToggleSetting("onLeft")
			end
		}

		local c = panel:NewCheckButton{
			name = "Collapse Down",
			get = function() return panel.owner.sets.revCollapse end,
			set = function(self, enable)
				panel.owner:ToggleSetting("revCollapse")
			end
		}

		local c = panel:NewCheckButton{
			name = "Enable Scroll",
			get = function() return panel.owner.sets.enableScroll end,
			set = function(self, enable)
				panel.owner:ToggleSetting("enableScroll")
			end
		}

		local c = panel:NewCheckButton{
			name = "Always Show Achievements",
			get = function() return panel.owner.sets.alwaysShowAchieves end,
			set = function(self, enable)
				panel.owner:ToggleSetting("alwaysShowAchieves")
			end
		}

		local slider =  panel:NewSlider({
			name = 'Achievement Height',
			min = 25,
			max = 350,
			get = function(self) --Getter
				return panel.owner.sets.achieveheight
			end,
			set = function(self) --Setter
				
				panel.owner:UpdateSetting("achieveheight", self:GetValue())

				return panel.owner.sets.achieveheight
			end,
		})
	end

	local function AddLayoutPanel(menu)
		local panel = menu:NewPanel(L.Layout)


		panel.scaleSlider = panel:NewScaleSlider()

		local slider =  panel:NewSlider({
			name = 'Height',
			min = 50,
			max = 900,
			get = function(self) --Getter
				return panel.owner.sets.height
			end,
			set = function(self) --Setter
				
				panel.owner:UpdateSetting("height", self:GetValue())

				return panel.owner.sets.height
			end,
		})
		
		panel.paddingSlider = panel:NewPaddingSlider()

		panel.opacitySlider = panel:NewOpacitySlider()
		panel.fadeSlider = panel:NewFadeSlider()
		return panel
	end

	local function AddShowState(self)
		local p = self:NewPanel(L.ShowStates)

		p.showStatesEditBox = p:NewTextInput{
			name = L.ShowStates,
			multiline = true,
			width = 290,
			height = 64,
			get = function() return p.owner:GetShowStates() end,
			set = function(self, value) p.owner:SetShowStates(value)end
		}

		return p
	end

	local function AddAdvancedPanel(menu)
		local panel = menu:NewPanel(L.Advanced)
		panel:NewClickThroughCheckbox()
		panel:NewShowInOverrideUICheckbox()
		panel:NewShowInPetBattleUICheckbox()
		return panel
	end

	function QuestBar:CreateMenu()
		local menu = Dominos:NewMenu(self.id)
		L = LibStub('AceLocale-3.0'):GetLocale('Dominos-Config')

		AddLayoutPanel(menu)
		AddQuestPanel(menu)
		AddShowState(menu)
		if self.components then
			for i, b in pairs(self.components) do
				if b.CreateMenu and QuestBar.Menu then
					b:CreateMenu(menu)
				end
			end
		end

		AddAdvancedPanel(menu)
		return menu
	end


	local menu, config
	function QuestBar:ShowMenu()
		if not QuestBar.master:IsConfigAddonEnabled() then return end
		config = config or LoadAddOn('Dominos_Config')
		L = L or LibStub('AceLocale-3.0'):GetLocale('Dominos-Config')
		QuestBar.menu = QuestBar.menu or self:CreateMenu()

		local menu = QuestBar.menu
		menu:Hide()
		menu:SetOwner(self)
		menu:ShowPanel(L.Layout)
		menu:Show()
	end
end






local op = {
	Left = "Right",
	Right = "Left",
	Bottom = "Top",
	Top = "Bottom",
}

function GameTooltip_AnchorComparisonTooltips(self, anchorFrame, shoppingTooltip1, shoppingTooltip2, primaryItemShown, secondaryItemShown)
	local leftPos = anchorFrame:GetLeft();
	local rightPos = anchorFrame:GetRight();

	local side;
	local anchorType = self:GetAnchorType();
	local totalWidth = 0;
	if ( primaryItemShown  ) then
		totalWidth = totalWidth + shoppingTooltip1:GetWidth();
	end
	if ( secondaryItemShown  ) then
		totalWidth = totalWidth + shoppingTooltip2:GetWidth();
	end
	if ( self.overrideComparisonAnchorSide ) then
		side = self.overrideComparisonAnchorSide;
	else
		-- find correct side
		local rightDist = 0;
		if ( not rightPos ) then
			rightPos = 0;
		end
		if ( not leftPos ) then
			leftPos = 0;
		end

		rightDist = GetScreenWidth() - rightPos;

		if ( anchorType and totalWidth < leftPos and (anchorType == "ANCHOR_LEFT" or anchorType == "ANCHOR_TOPLEFT" or anchorType == "ANCHOR_BOTTOMLEFT") ) then
			side = "left";
		elseif ( anchorType and totalWidth < rightDist and (anchorType == "ANCHOR_RIGHT" or anchorType == "ANCHOR_TOPRIGHT" or anchorType == "ANCHOR_BOTTOMRIGHT") ) then
			side = "right";
		elseif ( rightDist < leftPos ) then
			side = "left";
		else
			side = "right";
		end
	end

	-- see if we should slide the tooltip
	if ( anchorType and anchorType ~= "ANCHOR_PRESERVE" ) then
		if ( (side == "left") and (totalWidth > leftPos) ) then
			self:SetAnchorType(anchorType, (totalWidth - leftPos), 0);
		elseif ( (side == "right") and (rightPos + totalWidth) >  GetScreenWidth() ) then
			self:SetAnchorType(anchorType, -((rightPos + totalWidth) - GetScreenWidth()), 0);
		end
	end

	local _, anchor = anchorFrame:GetPoint()
	anchor = anchor or anchorFrame
	local x, y = anchor:GetCenter()


	if not (x or y) then
		if ( secondaryItemShown ) then
			shoppingTooltip2:SetOwner(self, "ANCHOR_NONE");
			shoppingTooltip2:ClearAllPoints();
			shoppingTooltip1:SetOwner(self, "ANCHOR_NONE");
			shoppingTooltip1:ClearAllPoints();

			if ( side and side == "left" ) then
				shoppingTooltip1:SetPoint("TOPRIGHT", anchorFrame, "TOPLEFT", 0, -10);
			else
				shoppingTooltip2:SetPoint("TOPLEFT", anchorFrame, "TOPRIGHT", 0, -10);
			end

			if ( side and side == "left" ) then
				shoppingTooltip2:SetPoint("TOPRIGHT", shoppingTooltip1, "TOPLEFT");
			else
				shoppingTooltip1:SetPoint("TOPLEFT", shoppingTooltip2, "TOPRIGHT");
			end
		else
			shoppingTooltip1:SetOwner(self, "ANCHOR_NONE");
			shoppingTooltip1:ClearAllPoints();

			if ( side and side == "left" ) then
				shoppingTooltip1:SetPoint("TOPRIGHT", anchorFrame, "TOPLEFT", 0, -10);
			else
				shoppingTooltip1:SetPoint("TOPLEFT", anchorFrame, "TOPRIGHT", 0, -10);
			end

			shoppingTooltip2:Hide();
		end

		return
	end

	local w, h = UIParent:GetSize()
	local half_w, half_h = w/2, h/2
	local ver, hor = "Top", "Left"
	if x >= half_w then
		hor = "Right"
	end


	if ( secondaryItemShown ) then
		shoppingTooltip2:SetOwner(self, "ANCHOR_NONE");
		shoppingTooltip2:ClearAllPoints();
		shoppingTooltip1:SetOwner(self, "ANCHOR_NONE");
		shoppingTooltip1:ClearAllPoints();

		if ( side and side == "left" ) then
			shoppingTooltip1:SetPoint(ver..hor, anchorFrame, ver..op[hor], 0, 0)
			shoppingTooltip2:SetPoint(ver..hor, shoppingTooltip1, ver..op[hor]);
		else
			shoppingTooltip2:SetPoint(ver..hor, anchorFrame, ver..op[hor], 0, 0)
			shoppingTooltip1:SetPoint(ver..hor, shoppingTooltip2, ver..op[hor]);
		end
	else
		shoppingTooltip1:SetOwner(self, "ANCHOR_NONE");
		shoppingTooltip1:ClearAllPoints();
		shoppingTooltip1:SetPoint(ver..hor, anchorFrame, ver..op[hor], 0, 0)
		shoppingTooltip2:Hide();
	end
end
